function yhat = exact_tree_proj(y,d,k)

%inputs: vector y, tree order d, sparsity k
%warning: y must be in R^N where N = d^J for some J

%calculate number of levels
N = length(y);
J = log2(N)/log2(d);

%initializations for level J
for i = d^(J-1)+1:d^J
F(i,1) = 0;
F(i,2) = [y(i)]^2;
end
lmax(J) = 1;

%proceed from fine to coarse
for j = J-1:-1:1

%calculate lmax(j)
lmax(j) = min((d^(J+1-j)-1)/(d-1),k-j);

%select node
for i = d^(j-1)+1:d^j

%set trivial optimal subtrees
F(i,1) = 0;
F(i,2) = (y(i))^2;
G(i,1,:) = zeros(d,1);
G(i,2,:) = zeros(d,1);

%initialize larger subtrees
for l = 2:lmax(j)
F(i,l+1) = 0;
G(i,l+1,:) = zeros(d,1);
end

lmax_child(j,1) = 1;

%incorporate children successively
for r = 1:d

lmax_child(j,r+1) = min(lmax(j),(r*lmax(j+1)+1));

%find all necessary optimal subtrees
for l = 2:lmax_child(j,r+1)

Ftemp(i,l+1) = F(i,l+1);
Gtemp(i,l+1,:) = G(i,l+1,:);

for s = max(1,l-lmax_child(j,r)):min(l-1,lmax(j+1))

Fmaxcand = F(i,l-s+1) + F(d*(i-1)+r,s+1);
if (Fmaxcand > Ftemp(i,l+1))
Ftemp(i,l+1) = Fmaxcand;
Gtemp(i,l+1,:) = G(i,l-s+1,:);
Gtemp(i,l+1,r) = s;
end

end

end

%store optimal results after each child is incorporated
for l = 2:lmax_child(j,r+1)

F(i,l+1) = Ftemp(i,l+1);
%F(i,l+1)
G(i,l+1,:) = zeros(d,1);
G(i,l+1,:) = Gtemp(i,l+1,:);

end

end

end

end

%find optimal subtrees for top node
F(1,1) = 0;
for l = 1:k
F(1,l+1) = (y(1))^2;
G(1,l+1,:) = zeros(d,1);
end

lmax_childk(1) = 1;

%incorporate children successively
for r = 2:d

lmax_childk(r) = min(k,(r-1)*lmax(1)+1);

%find all optimal subtrees up to k
for l = 2:lmax_childk(r)

Ftemp(1,l+1) = F(1,l+1);
Gtemp(1,l+1,:) = G(1,l+1,:);

for s = max(1,l-lmax_childk(r-1)):min(l-1,lmax(1))

Fmaxcand = F(1,l-s+1) + F(r,s+1);
if (Fmaxcand > Ftemp(1,l+1))
Ftemp(1,l+1) = Fmaxcand;
Gtemp(1,l+1,:) = G(1,l-s+1,:);
Gtemp(1,l+1,r) = s;
end

end

end

for l = 2:lmax_childk(r)

%store optimal results after each child is incorporated
F(1,l+1) = Ftemp(1,l+1);
G(1,l+1,:) = zeros(d,1);
G(1,l+1,:) = Gtemp(1,l+1,:);

end

end

%backtrack from coarse to fine to identify optimal tree

%initializations
yhat = zeros(N,1);
tau = zeros(N,1);
tau(1) = 1;
yhat(1,1) = y(1,1);
gamma(1) = k + 1;

%backtrack to level 1
for r = 2:d

if (G(1,k+1,r)>0)
tau(r) = 1;
yhat(r,1) = y(r,1);
gamma(r) = G(1,k+1,r);

else 
yhat(r,1) = 0;

end

end

%proceed from coarse to fine
for j = 1:J-1

for i = d^(j-1)+1:d^j

if (tau(i)==1)

for r = 1:d

if (G(i,gamma(i)+1,r)>0)
tau(d*(i-1)+r)=1;
yhat(d*(i-1)+r,1) = y(d*(i-1)+r,1);
gamma(d*(i-1)+r) = G(i,gamma(i)+1,r);

else
yhat(d*(i-1)+r,1) = 0;

end

end

end

end

end
